# -*- coding: utf-8 -*-

# Copyright (C) 2015 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Initializing and managing Gaupol windows."""

# For historical reasons gaupol.Application has been effectively
# a Gtk.Window. So, the later introduced Gtk.Application is for
# us a gaupol.ApplicationManager and it doesn't actually manage
# much, but is more of a main function in the form of a class.

import aeidon
import argparse
import gaupol
import os
import re
import sys

from aeidon.i18n   import _
from gi.repository import Gio
from gi.repository import GObject
from gi.repository import Gtk

__all__ = ("ApplicationManager",)


class ApplicationManager(Gtk.Application):

    """Initializing and managing Gaupol windows."""

    def __init__(self, args):
        """Initialize an :class:`ApplicationManager` instance."""
        GObject.GObject.__init__(self)
        self.menubar_builder = None
        self.set_application_id("io.otsaloma.gaupol")
        self.set_flags(Gio.ApplicationFlags.NON_UNIQUE)
        self.connect("activate", self._on_activate, args)
        self.connect("shutdown", self._on_shutdown)

    def _init_application(self, opts, args):
        """Initialize application and open files from `args`."""
        application = gaupol.Application()
        paths = list(map(os.path.abspath, args))
        application.open_main(paths, opts.encoding)
        page = application.get_current_page()
        if page is None: return
        if opts.translation_file is not None:
            path = os.path.abspath(opts.translation_file)
            method = opts.align_method.upper()
            method = getattr(aeidon.align_methods, method)
            application.open_translation(path, opts.encoding, method)
        if opts.video_file is not None:
            path = os.path.abspath(opts.video_file)
            page.project.video_path = path
            application.update_gui()
            application.load_video(path)
        if opts.jump is not None:
            page.view.set_focus(opts.jump)
            page.view.scroll_to_row(opts.jump)

    def _init_configuration(self):
        """Read configuration values from file."""
        gaupol.conf.path = os.path.join(
            aeidon.CONFIG_HOME_DIR, "gaupol.conf")
        gaupol.conf.read_from_file()
        if (gaupol.conf.general.dark_theme or
            os.getenv("GTK_THEME", "").endswith(":dark")):
            Gtk.Settings.get_default().set_property(
                "gtk-application-prefer-dark_theme", True)

    def _init_menubar(self):
        """Initialize the window menubar."""
        path = os.path.join(aeidon.DATA_DIR, "ui", "menubar.ui")
        self.menubar_builder = Gtk.Builder.new_from_file(path)
        self.set_menubar(self.menubar_builder.get_object("menubar"))

    def _on_activate(self, manager, args):
        """Initialize application and open files from `args`."""
        opts, args = self._parse_args(args)
        sys.excepthook = gaupol.util.show_exception
        self._init_configuration()
        self._init_menubar()
        self._init_application(opts, args)

    def _on_shutdown(self, manager):
        """Terminate application."""
        gaupol.conf.write_to_file()

    def _parse_args(self, args):
        """Parse and return options and arguments from `args`."""
        parser = argparse.ArgumentParser(
            usage=_("gaupol [OPTION...] [FILE...] [+[NUM]]"))

        parser.add_argument(
            "files",
            metavar=_("FILE..."),
            nargs="*",
            default=[],
            help=_("subtitle files to open"))

        parser.add_argument(
            "--version",
            action="version",
            version="gaupol {}".format(gaupol.__version__))

        parser.add_argument(
            "-e", "--encoding",
            action="store",
            metavar=_("ENCODING"),
            dest="encoding",
            default=None,
            help=_("set the character encoding used to open files"))

        parser.add_argument(
            "--list-encodings",
            action="store_true",
            dest="list_encodings",
            default=False,
            help=_("list all available character encodings"))

        parser.add_argument(
            "-t", "--translation-file",
            action="store",
            metavar=_("FILE"),
            dest="translation_file",
            default=None,
            help=_("open translation file"))

        parser.add_argument(
            "-a", "--align-method",
            action="store",
            metavar=_("METHOD"),
            dest="align_method",
            default="position",
            choices=["number", "position"],
            help=_("method used to align translation subtitles: 'number' or 'position'"))

        parser.add_argument(
            "-v", "--video-file",
            action="store",
            metavar=_("FILE"),
            dest="video_file",
            default=None,
            help=_("select video file"))

        args = parser.parse_args()
        args.jump = None
        if args.list_encodings:
            return self._print_encodings()
        if not args.encoding in (None, "auto"):
            args.encoding = aeidon.encodings.translate_code(args.encoding)
        # Parse row to jump to. If found, remove from args.files.
        for arg in filter(re.compile(r"\+\d*").match, args.files):
            args.jump = max(0, int(arg[1:]) - 1) if arg[1:] else -1
            args.files.remove(arg)
        return args, args.files

    def _print_encodings(self):
        """Print available character encodings and exit."""
        encodings = [x[0] for x in aeidon.encodings.get_valid()]
        if aeidon.util.chardet_available():
            encodings.insert(0, "auto")
        print(", ".join(encodings))
        raise SystemExit(0)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Dialog for selecting a subtitle file to save."""

import aeidon
import gaupol
import os

from aeidon.i18n   import _
from gi.repository import GObject
from gi.repository import Gtk

__all__ = ("SaveDialog",)


class SaveDialog(Gtk.FileChooserDialog, gaupol.FileDialog):

    """Dialog for selecting a subtitle file to save."""

    _widgets = [
        "encoding_combo",
        "format_combo",
        "framerate_combo",
        "framerate_label",
        "newline_combo",
    ]

    def __init__(self, parent, title, mode):
        """Initialize a :class:`SaveDialog` instance."""
        GObject.GObject.__init__(self)
        self._mode = mode
        self._init_dialog(parent, title)
        self._init_extra_widget()
        self._init_filters()
        self._init_format_combo()
        self._init_encoding_combo()
        self._init_newline_combo()
        self._init_framerate_combo()
        self._init_values()

    def get_format(self):
        """Return the selected format."""
        index = self._format_combo.get_active()
        return aeidon.formats[index]

    def get_framerate(self):
        """Return the selected framerate."""
        index = self._framerate_combo.get_active()
        return aeidon.framerates[index]

    def get_newline(self):
        """Return the selected newline."""
        index = self._newline_combo.get_active()
        return aeidon.newlines[index]

    def _init_dialog(self, parent, title):
        """Initialize the dialog."""
        self.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
        self.add_button(_("_Save"), Gtk.ResponseType.OK)
        self.set_default_response(Gtk.ResponseType.OK)
        self.set_transient_for(parent)
        self.set_title(title)
        self.connect("response", self._on_response)
        save_button = self.get_widget_for_response(Gtk.ResponseType.OK)
        save_button.connect("event", self._on_save_button_event)
        self.set_action(Gtk.FileChooserAction.SAVE)
        self.set_do_overwrite_confirmation(True)

    def _init_extra_widget(self):
        """Initialize the extra widget from UI definition file."""
        ui_file_path = os.path.join(aeidon.DATA_DIR, "ui", "save-dialog.ui")
        builder = Gtk.Builder()
        builder.set_translation_domain("gaupol")
        builder.add_from_file(ui_file_path)
        builder.connect_signals(self)
        for name in self._widgets:
            widget = builder.get_object(name)
            setattr(self, "_{}".format(name), widget)
        vbox = gaupol.util.new_vbox(spacing=0)
        main_vbox = builder.get_object("main_vbox")
        main_vbox.get_parent().remove(main_vbox)
        vbox.add(main_vbox)
        vbox.show_all()
        self.set_extra_widget(vbox)

    def _init_format_combo(self):
        """Initialize the format combo box."""
        store = Gtk.ListStore(str)
        self._format_combo.set_model(store)
        for name in (x.label for x in aeidon.formats):
            if name == "SubRip":
                # Mark the SubRip format as recommended, since
                # it's the most common one and best supported,
                # both in Gaupol and elsewhere.
                name = _("{} (recommended)").format(name)
            store.append((name,))
        view = self._format_combo.get_child()
        path = gaupol.util.tree_row_to_path(0)
        view.set_displayed_row(path)
        renderer = Gtk.CellRendererText()
        self._format_combo.pack_start(renderer, expand=True)
        self._format_combo.add_attribute(renderer, "text", 0)

    def _init_framerate_combo(self):
        """Initialize the framerate combo box."""
        store = Gtk.ListStore(str)
        self._framerate_combo.set_model(store)
        for name in (x.label for x in aeidon.framerates):
            store.append((name,))
        view = self._framerate_combo.get_child()
        path = gaupol.util.tree_row_to_path(0)
        view.set_displayed_row(path)
        renderer = Gtk.CellRendererText()
        self._framerate_combo.pack_start(renderer, expand=True)
        self._framerate_combo.add_attribute(renderer, "text", 0)

    def _init_newline_combo(self):
        """Initialize the newline combo box."""
        store = Gtk.ListStore(str)
        self._newline_combo.set_model(store)
        for name in (x.label for x in aeidon.newlines):
            store.append((name,))
        view = self._newline_combo.get_child()
        path = gaupol.util.tree_row_to_path(0)
        view.set_displayed_row(path)
        renderer = Gtk.CellRendererText()
        self._newline_combo.pack_start(renderer, expand=True)
        self._newline_combo.add_attribute(renderer, "text", 0)

    def _init_values(self):
        """Initialize default values for widgets."""
        if os.path.isdir(gaupol.conf.file.directory):
            self.set_current_folder(gaupol.conf.file.directory)
        self.set_encoding(gaupol.conf.file.encoding)
        self.set_format(gaupol.conf.file.format)
        self.set_newline(gaupol.conf.file.newline)
        self.set_framerate(gaupol.conf.editor.framerate)
        self._framerate_combo.hide()
        self._framerate_label.hide()

    def _on_format_combo_changed(self, *args):
        """Change the extension of the current filename."""
        format = self.get_format()
        path = self.get_filename()
        if path is not None:
            dirname = os.path.dirname(path)
            basename = os.path.basename(path)
            if not path.endswith(format.extension):
                basename = aeidon.util.replace_extension(basename, format)
                path = os.path.join(dirname, basename)
                self.unselect_filename(path)
                self.set_current_name(basename)
                self.set_filename(path)
        visible = (format.mode != self._mode)
        self._framerate_combo.set_visible(visible)
        self._framerate_label.set_visible(visible)

    def _on_response(self, dialog, response):
        """Save default values for widgets."""
        directory = self.get_current_folder()
        if directory is not None:
            gaupol.conf.file.directory = directory
        gaupol.conf.file.encoding = self.get_encoding()
        gaupol.conf.file.format = self.get_format()
        gaupol.conf.file.newline = self.get_newline()
        gaupol.conf.editor.framerate = self.get_framerate()
        if (response == Gtk.ResponseType.OK and
            not self.get_filename().endswith(self.get_format().extension)):
            # If the filename is lacking the extension, add it, stop this
            # response and emit a new one so that overwrite confirmation gets
            # called with the full filename. The filename extension might have
            # already been added in self._on_save_button_event, but not
            # necessarily if the user hit Enter on the keyboard.
            self._format_combo.emit("changed")
            gaupol.util.iterate_main()
            self.stop_emission("response")
            return self.response(Gtk.ResponseType.OK)

    def _on_save_button_event(self, button, event):
        """Ensure that the filename contains an extension."""
        # Add possibly lacking extension to the filename.
        self._format_combo.emit("changed")
        gaupol.util.iterate_main()

    def set_format(self, format):
        """Set the selected format."""
        if format is None: return
        self._format_combo.set_active(format)

    def set_framerate(self, framerate):
        """Set the selected framerate."""
        if framerate is None: return
        self._framerate_combo.set_active(framerate)

    def set_name(self, path):
        """Set the selected filename."""
        if os.path.isfile(path):
            return self.set_filename(path)
        return self.set_current_name(path)

    def set_newline(self, newline):
        """Set the selected newline."""
        if newline is None: return
        self._newline_combo.set_active(newline)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
Reading, writing and manipulating text-based subtitle files.

:mod:`aeidon` is a Python package that provides classes and functions for
dealing with text-based subtitle files of many different formats. Functions
exist for reading and writing subtitle files as well as manipulating subtitle
data, i.e. positions (times or frames) and texts. Two examples of basic use
of the :mod:`aeidon` package follow below.

Converting a file from the SubRip format to the MicroDVD format::

    project = aeidon.Project()
    project.open_main("/path/to/file.srt", "utf_8")
    project.set_framerate(aeidon.framerates.FPS_23_976)
    project.save_main(aeidon.files.new(aeidon.formats.MICRODVD,
                                       "/path/to/file.sub",
                                       "utf_8"))

Making all subtitles in a file appear two seconds earlier::

    project = aeidon.Project()
    project.open_main("/path/to/file.srt", "utf_8")
    project.shift_positions(None, aeidon.as_seconds(-2))
    project.save_main()

:mod:`aeidon` handles positions as either times (as strings of form
``HH:MM:SS.SSS``), frames (as integers) or in some cases seconds (as floats).
All these can be used with functions that edit subtitle data regardless of what
the native position type of the file format used is.

:mod:`aeidon` handles two separate documents that comprise a project -- a main
and a translation document. These correspond to separate files, but the
subtitles are common since positions are shared.

:mod:`aeidon` includes an undo/redo-system. Any subtitle data-editing methods
of :class:`aeidon.Project` (ones marked with the :func:`aeidon.deco.revertable`
decorator) can be undone and redone. If using :mod:`aeidon` in a context where
reverting actions is never needed, greater flexibility can be achieved by
accessing the subtitles directly (via :attr:`aeidon.Project.subtitles`).

:var CONFIG_HOME_DIR: Path to the user's local configuration directory
:var DATA_DIR: Path to the global data directory
:var DATA_HOME_DIR: Path to the user's local data directory
:var LOCALE_DIR: Path to the global locale directory
:var RE_ANY_TAG: Regular expression for markup tags of any format

:var align_methods: Enumerations for subtitle align methods
:var documents: Enumerations for document types
:var formats: Enumerations for subtitle file format types
:var framerates: Enumerations for framerate types
:var modes: Enumerations for position unit types
:var newlines: Enumerations for newline character types
:var players: Enumerations for video player application types
:var registers: Enumerations for action action reversion register types
"""

import re
import sys

__version__ = "1.10.1"

RUNNING_SPHINX = (sys.argv[0].endswith("autogen.py") or
                  sys.argv[0].endswith("sphinx-build"))

RE_ANY_TAG = re.compile(r"(^[/\\_]+|<.*?>|\{.*?\})")

try:
    import gi
    gi.require_version("Gspell", "1")
except Exception:
    pass

from aeidon.paths import * # noqa
from aeidon.position import * # noqa
from aeidon import deco # noqa
from aeidon import i18n # noqa
from aeidon import util # noqa
from aeidon import temp # noqa
from aeidon.delegate import * # noqa
from aeidon.singleton import * # noqa
from aeidon.mutables import * # noqa
from aeidon.observable import * # noqa
from aeidon.errors import * # noqa
from aeidon.enum import * # noqa
from aeidon.enums import * # noqa
from aeidon import encodings # noqa
from aeidon import languages # noqa
from aeidon import countries # noqa
from aeidon import locales # noqa
from aeidon import scripts # noqa
from aeidon.metadata import * # noqa
from aeidon.calculator import * # noqa
from aeidon.finder import * # noqa
from aeidon.parser import * # noqa
from aeidon.liner import * # noqa
from aeidon import containers # noqa
from aeidon.subtitle import * # noqa
from aeidon.file import * # noqa
from aeidon import files # noqa
from aeidon.markup import * # noqa
from aeidon import markups # noqa
from aeidon.markupconv import * # noqa
from aeidon.pattern import * # noqa
from aeidon.patternman import * # noqa
from aeidon.clipboard import * # noqa
from aeidon.revertable import * # noqa
from aeidon.spell import * # noqa
from aeidon import agents # noqa
from aeidon.project import * # noqa
from aeidon.unittest import * # noqa



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Internal text clipboard."""

__all__ = ("Clipboard",)


class Clipboard:

    """Internal text clipboard."""

    def __init__(self):
        """Initialize a :class:`Clipboard` instance."""
        # List of strings, each being the text of one subtitle. Nones in the
        # list express that those subtitles are skipped, i.e. the range of
        # subtitles is not unified, but contains gaps.
        self._texts = []

    def append(self, item):
        """Append `item` to texts."""
        self._texts.append(item)

    def clear(self):
        """Clear texts."""
        self._texts = []

    def get_string(self):
        """Return texts as one string."""
        strings = [x or "" for x in self._texts]
        return "\n\n".join(strings)

    def get_texts(self):
        """Return texts as a list of strings."""
        return self._texts[:]

    def is_empty(self):
        """Return ``True`` if empty."""
        return not bool(self._texts)

    def set_string(self, string):
        """Set the list of texts from a single string."""
        strings = [x or None for x in string.split("\n\n")]
        self.set_texts(strings)

    def set_texts(self, texts):
        """Set the list of texts."""
        self._texts = list(texts)



# -*- coding: utf-8 -*-

# Copyright (C) 2006 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Dialog for splitting a project in two."""

import aeidon
import gaupol

from aeidon.i18n   import _
from gi.repository import Gtk

__all__ = ("SplitDialog",)


class SplitDialog(gaupol.BuilderDialog):

    """Dialog for splitting a project in two."""

    _widgets = ["subtitle_spin"]

    def __init__(self, parent, application):
        """Initialize a :class:`SplitDialog` instance."""
        gaupol.BuilderDialog.__init__(self, "split-dialog.ui")
        self.application = application
        self._init_dialog(parent)
        self._init_subtitle_spin()

    def _init_dialog(self, parent):
        """Initialize the dialog."""
        self.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
        self.add_button(_("_Split"), Gtk.ResponseType.OK)
        self.set_default_response(Gtk.ResponseType.OK)
        self.set_transient_for(parent)
        self.set_modal(True)

    def _init_subtitle_spin(self):
        """Initialize the subtitle spin button."""
        page = self.application.get_current_page()
        rows = page.view.get_selected_rows()
        subtitle = rows[0] + 1 if rows else 2
        self._subtitle_spin.set_value(subtitle)
        self._subtitle_spin.set_range(2, len(page.project.subtitles))
        self._subtitle_spin.emit("value-changed")

    def _on_subtitle_spin_value_changed(self, spin_button):
        """Select matching row in the view."""
        page = self.application.get_current_page()
        row = self._subtitle_spin.get_value_as_int() - 1
        page.view.set_focus(row, None)

    def _on_response(self, dialog, response):
        """Split the current project if OK'd."""
        if response == Gtk.ResponseType.OK:
            self._split_project()

    def _remove_from_source(self, page, index):
        """Remove rows starting at `index` from `page`."""
        indices = list(range(index, len(page.project.subtitles)))
        page.project.block("action-done")
        page.project.remove_subtitles(indices)
        page.project.set_action_description(
            aeidon.registers.DO, _("Splitting project"))
        page.project.unblock("action-done")

    def _shift_destination(self, src, dst):
        """Shift subtitles in `dst` page."""
        amount = -1 * src.project.subtitles[-1].end_seconds
        dst.project.shift_positions(None, amount, register=None)

    def _split_project(self):
        """Split the current project in two."""
        gaupol.util.set_cursor_busy(self.application.window)
        index = self._subtitle_spin.get_value_as_int() - 1
        src = self.application.get_current_page()
        dst = gaupol.Page(next(self.application.counter))
        subtitles = [x.copy() for x in src.project.subtitles[index:]]
        indices = list(range(len(subtitles)))
        dst.project.insert_subtitles(indices, subtitles)
        dst.reload_view_all()
        self._remove_from_source(src, index)
        self._shift_destination(src, dst)
        self.application.add_page(dst)
        message = _('Split {amount:d} subtitles to project "{name}"').format(
            amount=len(dst.project.subtitles), name=dst.untitle)
        self.application.flash_message(message)
        gaupol.util.set_cursor_normal(self.application.window)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Editing subtitle data."""

import aeidon
import bisect
import gaupol

from aeidon.i18n import _, n_


class EditAgent(aeidon.Delegate):

    """Editing subtitle data."""

    def __init__(self, master):
        """Initialize an :class:`EditAgent` instance."""
        aeidon.Delegate.__init__(self, master)
        self._pref_dialog = None
        self._revert_in_progress = False

    @aeidon.deco.export
    def _on_clear_texts_activate(self, *args):
        """Clear the selected texts."""
        page = self.get_current_page()
        rows = page.view.get_selected_rows()
        col = page.view.get_focus()[1]
        doc = page.text_column_to_document(col)
        page.project.clear_texts(rows, doc)

    @aeidon.deco.export
    def _on_copy_texts_activate(self, *args):
        """Copy the selected texts to the clipboard."""
        page = self.get_current_page()
        rows = page.view.get_selected_rows()
        col = page.view.get_focus()[1]
        doc = page.text_column_to_document(col)
        page.project.copy_texts(rows, doc)
        self._sync_clipboards(page)

    @aeidon.deco.export
    def _on_cut_texts_activate(self, *args):
        """Cut the selected texts to the clipboard."""
        page = self.get_current_page()
        rows = page.view.get_selected_rows()
        col = page.view.get_focus()[1]
        doc = page.text_column_to_document(col)
        page.project.cut_texts(rows, doc)
        self._sync_clipboards(page)

    @aeidon.deco.export
    def _on_edit_next_value_activate(self, *args):
        """Edit the focused column of the next subtitle."""
        page = self.get_current_page()
        path, column = page.view.get_cursor()
        row = gaupol.util.tree_path_to_row(path)
        path = gaupol.util.tree_row_to_path(row + 1)
        page.view.set_cursor(path, column, True)

    @aeidon.deco.export
    def _on_edit_preferences_activate(self, *args):
        """Configure Gaupol."""
        if self._pref_dialog is not None:
            return self._pref_dialog.present()
        self._pref_dialog = gaupol.PreferencesDialog(self.window, self)
        aeidon.util.connect(self, "_pref_dialog", "response")
        self._pref_dialog.show()

    @aeidon.deco.export
    def _on_edit_value_activate(self, *args):
        """Edit the focused cell."""
        page = self.get_current_page()
        path, column = page.view.get_cursor()
        page.view.set_cursor(path, column, True)

    @aeidon.deco.export
    def _on_end_earlier_activate(self, *args):
        """End the selected subtitle earlier."""
        page = self.get_current_page()
        row = page.view.get_selected_rows()[0]
        pos = page.project.subtitles[row].end_seconds
        length = gaupol.conf.editor.stretch_length
        value = ((pos-0.001) // length) * length
        page.project.set_end(row, value)
        register = aeidon.registers.DO
        description = _("Stretching end position")
        page.project.set_action_description(register, description)
        # Group repeated stretches as one action.
        if (len(page.project.undoables) > 1 and
            page.project.undoables[1].description ==
            page.project.undoables[0].description):
            page.project.group_actions(register, 2, description)

    @aeidon.deco.export
    def _on_end_later_activate(self, *args):
        """End the selected subtitle later."""
        page = self.get_current_page()
        row = page.view.get_selected_rows()[0]
        pos = page.project.subtitles[row].end_seconds
        length = gaupol.conf.editor.stretch_length
        value = (((pos + 0.001) // length) + 1) * length
        page.project.set_end(row, value)
        register = aeidon.registers.DO
        description = _("Stretching end position")
        page.project.set_action_description(register, description)
        # Group repeated stretches as one action.
        if (len(page.project.undoables) > 1 and
            page.project.undoables[1].description ==
            page.project.undoables[0].description):
            page.project.group_actions(register, 2, description)

    @aeidon.deco.export
    def _on_extend_selection_to_beginning_activate(self, *args):
        """Extend the selection up to the first subtitle."""
        page = self.get_current_page()
        row = page.view.get_selected_rows()[-1]
        rows = list(range(0, row + 1))
        page.view.select_rows(rows)

    @aeidon.deco.export
    def _on_extend_selection_to_end_activate(self, *args):
        """Extend the selection up to the last subtitle."""
        page = self.get_current_page()
        row = page.view.get_selected_rows()[0]
        rows = list(range(row, len(page.project.subtitles)))
        page.view.select_rows(rows)

    @aeidon.deco.export
    def _on_insert_subtitle_at_video_position_activate(self, *args):
        """Insert a new subtitle at video position."""
        mode = aeidon.modes.SECONDS
        pos = self.player.get_position(mode)
        if pos is None: return
        page = self.get_current_page()
        starts = [x.start_seconds for x in page.project.subtitles]
        index = bisect.bisect_right(starts, pos)
        subtitle = page.project.new_subtitle()
        subtitle.start_seconds = pos
        subtitle.end_seconds = pos + 3.0
        subtitle.main_text = "[{:d}]".format(index + 1)
        page.project.insert_subtitles((index,), (subtitle,))

    @aeidon.deco.export
    def _on_insert_subtitles_activate(self, *args):
        """Insert subtitles."""
        dialog = gaupol.InsertDialog(self.window, self)
        gaupol.util.flash_dialog(dialog)

    @aeidon.deco.export
    def _on_invert_selection_activate(self, *args):
        """Invert the current selection."""
        page = self.get_current_page()
        rows = set(range(0, len(page.project.subtitles)))
        rows -= set(page.view.get_selected_rows())
        page.view.select_rows(rows)

    @aeidon.deco.export
    def _on_merge_subtitles_activate(self, *args):
        """Merge the selected subtitles."""
        page = self.get_current_page()
        rows = page.view.get_selected_rows()
        page.project.merge_subtitles(rows)

    @aeidon.deco.export
    def _on_paste_texts_activate(self, *args):
        """Paste texts from the clipboard."""
        page = self.get_current_page()
        text = self.x_clipboard.wait_for_text()
        if text:
            # Update all clipboards in case text is being
            # copied from the external clipboard.
            page.project.clipboard.set_string(text)
            self._sync_clipboards(page)
        if page.project.clipboard.is_empty(): return
        rows = page.view.get_selected_rows()
        row, col = page.view.get_focus()
        doc = page.text_column_to_document(col)
        length = len(page.project.subtitles)
        # Ensure that even if new subtitles need to be inserted,
        # focus and scroll position are not moved to the end.
        rect = page.view.get_visible_rect()
        window = page.view.get_bin_window()
        window.freeze_updates()
        rows = page.project.paste_texts(rows[0], doc)
        rows = page.view.get_selected_rows()
        page.view.set_focus(row, col)
        page.view.select_rows(rows)
        page.view.scroll_to_point(rect.x, rect.y)
        window.thaw_updates()
        count = len(page.project.subtitles) - length
        if count <= 0: return
        self.flash_message(n_(
            "Inserted {:d} subtitle to fit clipboard contents",
            "Inserted {:d} subtitles to fit clipboard contents",
            count).format(count))

    def _on_pref_dialog_response(self, *args):
        """Destroy the preferences dialog."""
        self._pref_dialog.destroy()
        self._pref_dialog = None
        gaupol.conf.write_to_file()

    @aeidon.deco.export
    def _on_project_action_done(self, *args):
        """Update user interface and send a signal."""
        page = self.get_current_page()
        self.update_gui()
        self.emit("page-changed", page)

    @aeidon.deco.export
    def _on_project_action_redone(self, *args):
        """Update user interface and send a signal."""
        page = self.get_current_page()
        row = page.view.get_focus()[0]
        if row is not None:
            page.view.scroll_to_row(row)
        self.update_gui()
        self.emit("page-changed", page)

    @aeidon.deco.export
    def _on_project_action_undone(self, *args):
        """Update user interface and send a signal."""
        page = self.get_current_page()
        row = page.view.get_focus()[0]
        if row is not None:
            page.view.scroll_to_row(row)
        self.update_gui()
        self.emit("page-changed", page)

    @aeidon.deco.export
    def _on_redo_action_activate(self, *args):
        """Redo the last undone action."""
        self.redo()

    @aeidon.deco.export
    def _on_remove_subtitles_activate(self, *args):
        """Remove the selected subtitles."""
        page = self.get_current_page()
        rows = page.view.get_selected_rows()
        page.project.remove_subtitles(rows)

    @aeidon.deco.export
    def _on_select_all_activate(self, *args):
        """Select all subtitles."""
        page = self.get_current_page()
        selection = page.view.get_selection()
        selection.select_all()

    @aeidon.deco.export
    def _on_select_previous_from_video_position_activate(self, *args):
        """Select the previous subtitle from video position."""
        mode = aeidon.modes.SECONDS
        pos = self.player.get_position(mode)
        if pos is None: return
        page = self.get_current_page()
        for i, subtitle in reversed(list(enumerate(page.project.subtitles))):
            if subtitle.start_seconds < pos:
                page.view.select_rows([i])
                page.view.scroll_to_row([i])
                return
        # Select the first subtitle, if nothing previous.
        page.view.select_rows([0])
        page.view.scroll_to_row([0])

    @aeidon.deco.export
    def _on_select_next_from_video_position_activate(self, *args):
        """Select the next subtitle from video position."""
        mode = aeidon.modes.SECONDS
        pos = self.player.get_position(mode)
        if pos is None: return
        page = self.get_current_page()
        for i, subtitle in enumerate(page.project.subtitles):
            if subtitle.start_seconds > pos:
                page.view.select_rows([i])
                page.view.scroll_to_row([i])
                return
        # Select the last subtitle, if nothing next.
        last = len(page.project.subtitles) - 1
        page.view.select_rows([last])
        page.view.scroll_to_row([last])

    @aeidon.deco.export
    def _on_set_end_from_video_position_activate(self, *args):
        """Set subtitle end from video position."""
        page = self.get_current_page()
        mode = aeidon.modes.SECONDS
        row = page.view.get_selected_rows()[0]
        pos = self.player.get_position(mode)
        if pos is None: return
        page.project.set_end(row, pos)

    @aeidon.deco.export
    def _on_set_start_from_video_position_activate(self, *args):
        """Set subtitle start from video position."""
        page = self.get_current_page()
        mode = aeidon.modes.SECONDS
        row = page.view.get_selected_rows()[0]
        pos = self.player.get_position(mode)
        if pos is None: return
        page.project.set_start(row, pos)

    @aeidon.deco.export
    def _on_split_subtitle_activate(self, *args):
        """Split the selected subtitle."""
        page = self.get_current_page()
        row = page.view.get_selected_rows()[0]
        page.project.split_subtitle(row)

    @aeidon.deco.export
    def _on_start_earlier_activate(self, *args):
        """Start the selected subtitle earlier."""
        page = self.get_current_page()
        row = page.view.get_selected_rows()[0]
        pos = page.project.subtitles[row].start_seconds
        length = gaupol.conf.editor.stretch_length
        value = ((pos-0.001) // length) * length
        page.project.set_start(row, value)
        register = aeidon.registers.DO
        description = _("Stretching start position")
        page.project.set_action_description(register, description)
        # Group repeated stretches as one action.
        if (len(page.project.undoables) > 1 and
            page.project.undoables[1].description ==
            page.project.undoables[0].description):
            page.project.group_actions(register, 2, description)

    @aeidon.deco.export
    def _on_start_later_activate(self, *args):
        """Start the selected subtitle later."""
        page = self.get_current_page()
        row = page.view.get_selected_rows()[0]
        pos = page.project.subtitles[row].start_seconds
        length = gaupol.conf.editor.stretch_length
        value = (((pos + 0.001) // length) + 1) * length
        page.project.set_start(row, value)
        register = aeidon.registers.DO
        description = _("Stretching start position")
        page.project.set_action_description(register, description)
        # Group repeated stretches as one action.
        if (len(page.project.undoables) > 1 and
            page.project.undoables[1].description ==
            page.project.undoables[0].description):
            page.project.group_actions(register, 2, description)

    @aeidon.deco.export
    def _on_undo_action_activate(self, *args):
        """Undo the last action."""
        self.undo()

    @aeidon.deco.export
    def _on_view_renderer_edited(self, renderer, path, value, column):
        """Save changes made while editing cell."""
        self._set_unsafe_enabled(True)
        self.show_message(None)
        page = self.get_current_page()
        row = gaupol.util.tree_path_to_row(path)
        col = page.view.get_columns().index(column)
        if page.view.is_position_column(col):
            if not value: return
            if page.edit_mode == aeidon.modes.FRAME:
                with aeidon.util.silent(ValueError):
                    value = aeidon.as_frame(value)
        if col == page.view.columns.START:
            return page.project.set_start(row, value)
        if col == page.view.columns.END:
            return page.project.set_end(row, value)
        if col ==  page.view.columns.DURATION:
            if page.edit_mode == aeidon.modes.TIME:
                value = value.replace(",", ".")
                with aeidon.util.silent(ValueError):
                    value = aeidon.as_seconds(value)
            return page.project.set_duration(row, value)
        doc = page.text_column_to_document(col)
        page.project.set_text(row, doc, value)
        self.update_gui()

    @aeidon.deco.export
    def _on_view_renderer_editing_canceled(self, *args):
        """Unset state set for editing cell."""
        self._set_unsafe_enabled(True)
        self.show_message(None)
        self.update_gui()

    @aeidon.deco.export
    def _on_view_renderer_editing_started(self, renderer, editor, path, column):
        """Set proper state for editing cell."""
        self._set_unsafe_enabled(False)
        page = self.get_current_page()
        col = page.view.get_columns().index(column)
        if not page.view.is_text_column(col): return
        self._set_global_italicize(page, col)
        start, end = page.view.get_visible_range()
        end = gaupol.util.tree_path_to_row(end)
        if gaupol.util.tree_path_to_row(path) < end - 1:
            # Don't show help text if it would overlap with text being edited.
            self.show_message(_("Use Shift+Return for line-break"))

    @aeidon.deco.export
    def redo(self, count=1):
        """Redo `count` amount of actions."""
        # Avoid problems if holding down Ctrl+Shift+Z.
        if self._revert_in_progress: return
        self._revert_in_progress = True
        try:
            gaupol.util.set_cursor_busy(self.window)
            page = self.get_current_page()
            # Make sure the UI is given time to update
            # after the previous redone action.
            gaupol.util.iterate_main()
            if page.project.can_redo(count):
                page.project.redo(count)
            gaupol.util.set_cursor_normal(self.window)
        finally:
            self._revert_in_progress = False

    def _set_global_italicize(self, page, col):
        """Set global function to italicize text."""
        # MultilineCellRenderer's CellTextView needs to know how
        # to italicize text. Pass ugly via a global function.
        gaupol.italic_tag = None
        gaupol.italicize = None
        if col is None: return
        if not page.view.is_text_column(col): return
        doc = page.text_column_to_document(col)
        markup = page.project.get_markup(doc)
        if markup is None: return
        if markup.italic_tag is None: return
        gaupol.italic_tag = markup.italic_tag
        gaupol.italicize = markup.italicize

    def _set_unsafe_enabled(self, enabled):
        """Set enabled states of unsafe actions."""
        for name in self.window.list_actions():
            action = self.get_action(name)
            if action.action_group == "unsafe":
                action.set_enabled(enabled)

    def _sync_clipboards(self, page):
        """Synchronize all clipboards to match that of `page`."""
        texts = page.project.clipboard.get_texts()
        self.clipboard.set_texts(texts)
        for item in self.pages:
            item.project.clipboard.set_texts(texts)
        text = page.project.clipboard.get_string()
        self.x_clipboard.set_text(text, -1)
        self.update_gui()

    @aeidon.deco.export
    def undo(self, count=1):
        """Undo `count` amount of actions."""
        # Avoid problems if holding down Ctrl+Z.
        if self._revert_in_progress: return
        self._revert_in_progress = True
        try:
            gaupol.util.set_cursor_busy(self.window)
            page = self.get_current_page()
            # Make sure the UI is given time to update
            # after the previous undone action.
            gaupol.util.iterate_main()
            if page.project.can_undo(count):
                page.project.undo(count)
            gaupol.util.set_cursor_normal(self.window)
        finally:
            self._revert_in_progress = False



# -*- coding: utf-8 -*-

# Copyright (C) 2016 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""CSS styles and helper functions for styling."""

import aeidon
import gaupol
import os

from gi.repository import Gtk
from gi.repository import Pango

CSS_FILE = os.path.join(aeidon.DATA_DIR, "ui", "gaupol.css")
with open(CSS_FILE, "r") as f:
    CSS = f.read()


def _get_editor_font_css():
    """Return CSS for custom editor font."""
    font_desc = Pango.FontDescription("monospace")
    if (gaupol.conf.editor.custom_font and
        gaupol.conf.editor.use_custom_font):
        font_desc = Pango.FontDescription(gaupol.conf.editor.custom_font)
    # They fucking broke theming again with GTK 3.22.
    unit = "pt" if Gtk.check_version(3, 22, 0) is None else "px"
    css = """
    .gaupol-custom-font {{
        font-family: {family},monospace;
        font-size: {size}{unit};
        font-weight: {weight};
    }}""".format(
        family=font_desc.get_family().split(",")[0],
        size=int(round(font_desc.get_size() / Pango.SCALE)),
        unit=unit,
        weight=int(font_desc.get_weight()))
    css = css.replace("font-size: 0{unit};".format(unit=unit), "")
    css = css.replace("font-weight: 0;", "")
    css = "\n".join(filter(lambda x: x.strip(), css.splitlines()))
    return css

def load_css(widget):
    """Load CSS rules from file and conf for `widget`."""
    provider = Gtk.CssProvider()
    css = "\n".join((CSS, _get_editor_font_css()))
    provider.load_from_data(bytes(css.encode()))
    style = widget.get_style_context()
    priority = Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
    style.add_provider_for_screen(widget.get_screen(),
                                  provider,
                                  priority)

def _update_css(*args, **kwargs):
    """Update CSS rules for the default provider."""
    provider = Gtk.CssProvider()
    provider.load_from_data(bytes(_get_editor_font_css().encode()))

def use_font(widget, font):
    """Use `font` ("custom" or "monospace") for `widget`."""
    if not font: return
    load_css(widget)
    style = widget.get_style_context()
    style.add_class({
        "custom": "gaupol-custom-font",
        "monospace": "monospace",
    }[font])


gaupol.conf.editor.connect("notify::custom_font", _update_css)
gaupol.conf.editor.connect("notify::use_custom_font", _update_css)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Time and frame calculator."""

import aeidon

__all__ = ("Calculator",)


class Calculator:

    """
    Time and frame calculator.

    Times are handled as strings, frames as integers and seconds as floats.
    Only one instance of :class:`Calculator` exists for a given framerate.
    """

    _instances = {}

    def __new__(cls, framerate=None):
        """
        Return possibly existing instance for `framerate`.

        `framerate` can be either an :attr:`aeidon.framerates` item
        (preferred to be able to reuse instances) or an exact float value.
        """
        if framerate is None:
            framerate = aeidon.framerates.FPS_23_976
        if not framerate in aeidon.framerates:
            # Always return a new instance for float values.
            return object.__new__(cls)
        if not framerate in cls._instances:
            cls._instances[framerate] = object.__new__(cls)
        return cls._instances[framerate]

    def __init__(self, framerate=None):
        """
        Initialize a :class:`Calculator` instance.

        `framerate` can be either an :attr:`aeidon.framerates` item
        (preferred to be able to reuse instances) or an exact float value.
        """
        if framerate is None:
            framerate = aeidon.framerates.FPS_23_976
        if framerate in aeidon.framerates:
            self._framerate = framerate.value
        else:
            # Use non-constant values as is.
            self._framerate = float(framerate)

    def add(self, x, y):
        """Add position `y` to `x`."""
        if aeidon.is_time(x):
            x = self.to_seconds(x)
            y = self.to_seconds(y)
            return self.seconds_to_time(x + y)
        if aeidon.is_frame(x):
            return x + self.to_frame(y)
        if aeidon.is_seconds(x):
            return x + self.to_seconds(y)
        raise ValueError("Invalid type for x: {!r}"
                         .format(type(x)))

    def frame_to_seconds(self, frame):
        """Convert `frame` to seconds."""
        return aeidon.as_seconds(frame / self._framerate)

    def frame_to_time(self, frame):
        """Convert `frame` to time."""
        seconds = self.frame_to_seconds(frame)
        return self.seconds_to_time(seconds)

    def get_middle(self, x, y):
        """Return time, frame or seconds halfway between `x` and `y`."""
        if aeidon.is_time(x):
            x = self.time_to_seconds(x)
            y = self.to_seconds(y)
            return self.seconds_to_time((x + y) / 2)
        if aeidon.is_frame(x):
            y = self.to_frame(y)
            return aeidon.as_frame(round((x + y) / 2, 0))
        if aeidon.is_seconds(x):
            y = self.to_seconds(y)
            return aeidon.as_seconds(((x + y) / 2))
        raise ValueError("Invalid type for x: {!r}"
                         .format(type(x)))

    def is_earlier(self, x, y):
        """Return ``True`` if `x` is earlier than `y`."""
        if aeidon.is_time(x):
            x = self.time_to_seconds(x)
            return self.is_earlier(x, y)
        if aeidon.is_frame(x):
            return (x < self.to_frame(y))
        if aeidon.is_seconds(x):
            return (x < self.to_seconds(y))
        raise ValueError("Invalid type for x: {!r}"
                         .format(type(x)))

    def is_later(self, x, y):
        """Return ``True`` if `x` is later than `y`."""
        if aeidon.is_time(x):
            x = self.time_to_seconds(x)
            return self.is_later(x, y)
        if aeidon.is_frame(x):
            return (x > self.to_frame(y))
        if aeidon.is_seconds(x):
            return (x > self.to_seconds(y))
        raise ValueError("Invalid type for x: {!r}"
                         .format(type(x)))

    def is_valid_time(self, time):
        """Return ``True`` if `time` is a valid time string."""
        if time.startswith("-"):
            time = time[1:]
        try:
            hours    = int(time[ :2])
            minutes  = int(time[3:5])
            seconds  = int(time[6:8])
            mseconds = int(time[9: ])
        except ValueError:
            return False
        return (0 <= hours    <=  99 and
                0 <= minutes  <=  59 and
                0 <= seconds  <=  59 and
                0 <= mseconds <= 999)

    def normalize_time(self, time):
        """
        Convert `time` to valid format.

        >>> calc = aeidon.Calculator()
        >>> calc.normalize_time("1:2:3,4")
        '01:02:03.400'
        """
        time = time.strip()
        sign = "-" if time.startswith("-") else ""
        time = time.replace("-", "")
        time = time.replace(",", ".")
        if time.count(":") == 1:
            # Allow hours to be missing,
            # used at least in the WebVTT format.
            time = "00:{}".format(time)
        hours, minutes, seconds = time.split(":")
        return ("{}{:02.0f}:{:02.0f}:{:02.0f}.{:03.0f}"
                .format(sign,
                        int(hours),
                        int(minutes),
                        int(float(seconds)),
                        (float(seconds) % 1) * 1000))

    def round(self, pos, ndigits):
        """
        Round `pos` to given precision in decimal digits.

        `ndigits` may be negative. For frames zero will be used
        if given `ndigits` is greater than zero.
        """
        if aeidon.is_time(pos):
            pos = self.time_to_seconds(pos)
            pos = round(pos, ndigits)
            return self.seconds_to_time(pos)
        if aeidon.is_frame(pos):
            ndigits = min(0, ndigits)
            pos = round(pos, ndigits)
            return aeidon.as_frame(pos)
        if aeidon.is_seconds(pos):
            pos = round(pos, ndigits)
            return aeidon.as_seconds(pos)
        raise ValueError("Invalid type for pos: {!r}"
                         .format(type(pos)))

    def seconds_to_frame(self, seconds):
        """Convert `seconds` to frame."""
        return int(round(seconds * self._framerate, 0))

    def seconds_to_time(self, seconds):
        """Convert `seconds` to time."""
        sign = "-" if seconds < 0 else ""
        seconds = abs(round(seconds, 3))
        if seconds > 359999.999:
            return "{}99:59:59.999".format(sign)
        return ("{}{:02.0f}:{:02.0f}:{:02.0f}.{:03.0f}"
                .format(sign,
                        seconds // 3600,
                        (seconds % 3600) // 60,
                        int(seconds % 60),
                        (seconds % 1) * 1000))

    def time_to_frame(self, time):
        """Convert `time` to frame."""
        seconds = self.time_to_seconds(time)
        return self.seconds_to_frame(seconds)

    def time_to_seconds(self, time):
        """Convert `time` to seconds."""
        coefficient = -1 if time.startswith("-") else 1
        time = time[1:] if time.startswith("-") else time
        return coefficient * sum((float(time[ :2]) * 3600,
                                  float(time[3:5]) * 60,
                                  float(time[6:8]),
                                  float(time[9: ]) / 1000))

    def to_frame(self, pos):
        """Convert `pos` to frame."""
        if aeidon.is_time(pos):
            return self.time_to_frame(pos)
        if aeidon.is_frame(pos):
            return pos
        if aeidon.is_seconds(pos):
            return self.seconds_to_frame(pos)
        raise ValueError("Invalid type for pos: {!r}"
                         .format(type(pos)))

    def to_seconds(self, pos):
        """Convert `pos` to seconds."""
        if aeidon.is_time(pos):
            return self.time_to_seconds(pos)
        if aeidon.is_frame(pos):
            return self.frame_to_seconds(pos)
        if aeidon.is_seconds(pos):
            return pos
        raise ValueError("Invalid type for pos: {!r}"
                         .format(type(pos)))

    def to_time(self, pos):
        """Convert `pos` to time."""
        if aeidon.is_time(pos):
            return pos
        if aeidon.is_frame(pos):
            return self.frame_to_time(pos)
        if aeidon.is_seconds(pos):
            return self.seconds_to_time(pos)
        raise ValueError("Invalid type for pos: {!r}"
                         .format(type(pos)))



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Dialog for selecting properties to save multiple files with."""

import aeidon
import gaupol
import os

from aeidon.i18n   import _
from gi.repository import Gtk

__all__ = ("MultiSaveDialog",)


class MultiSaveDialog(gaupol.BuilderDialog, gaupol.FileDialog):

    """Dialog for selecting properties to save multiple files with."""

    _widgets = [
        "encoding_combo",
        "filechooser_button",
        "format_combo",
        "framerate_combo",
        "framerate_label",
        "newline_combo",
    ]

    def __init__(self, parent, application, modes):
        """Initialize a :class:`MultiSaveDialog` instance."""
        gaupol.BuilderDialog.__init__(self, "multi-save-dialog.ui")
        self.application = application
        self._modes = modes
        self._init_dialog(parent)
        self._init_format_combo()
        self._init_encoding_combo()
        self._init_newline_combo()
        self._init_framerate_combo()
        width = gaupol.util.char_to_px(60)
        self._filechooser_button.set_size_request(width, -1)
        self._init_values()

    def get_directory(self):
        """Return the selected directory."""
        return self._filechooser_button.get_filename()

    def get_format(self):
        """Return the selected format."""
        index = self._format_combo.get_active()
        return aeidon.formats[index]

    def get_framerate(self):
        """Return the selected framerate."""
        index = self._framerate_combo.get_active()
        return aeidon.framerates[index]

    def get_newline(self):
        """Return the selected newline."""
        index = self._newline_combo.get_active()
        return aeidon.newlines[index]

    def _get_pages(self):
        """Return a list of pages to consider."""
        return [x for x in self.application.pages
                if x.project.main_file is not None]

    def _init_dialog(self, parent):
        """Initialize the dialog."""
        self.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
        self.add_button(_("_Save"), Gtk.ResponseType.OK)
        self.set_default_response(Gtk.ResponseType.OK)
        self.set_transient_for(parent)
        self.set_modal(True)

    def _init_format_combo(self):
        """Initialize the format combo box."""
        store = Gtk.ListStore(str)
        self._format_combo.set_model(store)
        for name in (x.label for x in aeidon.formats):
            if name == "SubRip":
                # Mark the SubRip format as recommended, since
                # it's the most common one and best supported,
                # both in Gaupol and elsewhere.
                name = _("{} (recommended)").format(name)
            store.append((name,))
        view = self._format_combo.get_child()
        view.set_displayed_row(gaupol.util.tree_row_to_path(0))
        renderer = Gtk.CellRendererText()
        self._format_combo.pack_start(renderer, expand=True)
        self._format_combo.add_attribute(renderer, "text", 0)

    def _init_framerate_combo(self):
        """Initialize the framerate combo box."""
        store = Gtk.ListStore(str)
        self._framerate_combo.set_model(store)
        for name in (x.label for x in aeidon.framerates):
            store.append((name,))
        view = self._framerate_combo.get_child()
        path = gaupol.util.tree_row_to_path(0)
        view.set_displayed_row(path)
        renderer = Gtk.CellRendererText()
        self._framerate_combo.pack_start(renderer, expand=True)
        self._framerate_combo.add_attribute(renderer, "text", 0)

    def _init_newline_combo(self):
        """Initialize the newline combo box."""
        store = Gtk.ListStore(str)
        self._newline_combo.set_model(store)
        for name in (x.label for x in aeidon.newlines):
            store.append((name,))
        view = self._newline_combo.get_child()
        view.set_displayed_row(gaupol.util.tree_row_to_path(0))
        renderer = Gtk.CellRendererText()
        self._newline_combo.pack_start(renderer, expand=True)
        self._newline_combo.add_attribute(renderer, "text", 0)

    def _init_values(self):
        """Initialize default values for widgets."""
        # Suggest format, encoding and newlines to match
        # those from the first page to consider.
        pages = self._get_pages()
        file = pages[0].project.main_file
        self.set_directory(os.path.dirname(file.path))
        self.set_format(file.format)
        self.set_encoding(file.encoding)
        self.set_newline(file.newline)
        self.set_framerate(gaupol.conf.editor.framerate)
        visible = len(set(self._modes)) > 1
        self._framerate_combo.set_visible(visible)
        self._framerate_label.set_visible(visible)

    def _on_format_combo_changed(self, *args):
        """Change the extension of the current filename."""
        format = self.get_format()
        modes = list(self._modes) + [format.mode]
        visible = len(set(modes)) > 1
        self._framerate_combo.set_visible(visible)
        self._framerate_label.set_visible(visible)

    def _on_response(self, dialog, response):
        """Save default values for widgets."""
        gaupol.conf.file.encoding = self.get_encoding()
        gaupol.conf.file.format = self.get_format()
        gaupol.conf.file.newline = self.get_newline()
        gaupol.conf.editor.framerate = self.get_framerate()
        if response != Gtk.ResponseType.OK: return
        gaupol.util.set_cursor_busy(self._dialog)
        self._save_all_documents_as()
        gaupol.util.set_cursor_normal(self._dialog)

    @aeidon.deco.silent(gaupol.Default)
    def _save_all_documents_as(self):
        """Save all documents with selected properties."""
        pages = self._get_pages()
        files = [None for x in pages]
        for i, page in enumerate(pages):
            if self._framerate_combo.get_visible():
                # Set framerate to the selected one.
                framerate = self.get_framerate()
                self.application.set_current_page(page)
                action = self.application.get_action("set-framerate")
                action.activate(str(framerate))
            path = os.path.basename(page.project.main_file.path)
            path = aeidon.util.replace_extension(path, self.get_format())
            path = os.path.join(self.get_directory(), path)
            files[i] = aeidon.files.new(self.get_format(),
                                        path,
                                        self.get_encoding(),
                                        self.get_newline())

        if sum(os.path.isfile(x.path) for x in files) > 0:
            self._show_overwrite_question_dialog(files, self.get_directory())
        for i, page in enumerate(pages):
            self.application.save_main_as(page, files[i])

    def set_directory(self, path):
        """Set the selected directory."""
        return self._filechooser_button.set_filename(path)

    def set_format(self, format):
        """Set the selected format."""
        if format is None: return
        self._format_combo.set_active(format)

    def set_framerate(self, framerate):
        """Set the selected framerate."""
        if framerate is None: return
        self._framerate_combo.set_active(framerate)

    def set_newline(self, newline):
        """Set the selected newline."""
        if newline is None: return
        self._newline_combo.set_active(newline)

    def _show_overwrite_question_dialog(self, files, path):
        """Show a question dialog if about to overwrite files."""
        n = sum(os.path.isfile(x.path) for x in files)
        title = _("{:d} of the files to be saved already exist. Do you want to replace them?").format(n)
        message = _('The files already exist in "{}". Replacing them will overwrite their contents.').format(path)
        dialog = gaupol.QuestionDialog(self._dialog, title, message)
        dialog.add_button(_("_Cancel"), Gtk.ResponseType.NO)
        dialog.add_button(_("_Replace"), Gtk.ResponseType.YES)
        dialog.set_default_response(Gtk.ResponseType.YES)
        response = gaupol.util.flash_dialog(dialog)
        gaupol.util.raise_default(response != Gtk.ResponseType.YES)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""View actions for :class:`gaupol.Application`."""

import aeidon
import gaupol

class SetEditModeAction(gaupol.RadioAction):
    def __new__(cls):
        action = gaupol.RadioAction.new("set-edit-mode")
        action.__class__ = cls
        return action
    def __init__(self):
        gaupol.RadioAction.__init__(self, "set-edit-mode")
        self.accelerators = ["<Shift>T"]
        self.set_state(str(gaupol.conf.editor.mode))
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)

class SetFramerateAction(gaupol.RadioAction):
    def __new__(cls):
        action = gaupol.RadioAction.new("set-framerate")
        action.__class__ = cls
        return action
    def __init__(self):
        gaupol.RadioAction.__init__(self, "set-framerate")
        self.action_group = "unsafe"
        self.set_state(str(gaupol.conf.editor.framerate))
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(page.project.main_file is not None)

class SetLayoutAction(gaupol.RadioAction):
    def __new__(cls):
        action = gaupol.RadioAction.new("set-layout")
        action.__class__ = cls
        return action
    def __init__(self):
        gaupol.RadioAction.__init__(self, "set-layout")
        self.action_group = "unsafe"
        self.set_state(str(gaupol.conf.application_window.layout))
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(application.player is not None)

class ToggleDurationColumnAction(gaupol.ToggleAction):
    def __new__(cls):
        action = gaupol.ToggleAction.new("toggle-duration-column")
        action.__class__ = cls
        return action
    def __init__(self):
        gaupol.ToggleAction.__init__(self, "toggle-duration-column")
        self.action_group = "unsafe"
        self.set_state(gaupol.fields.DURATION in gaupol.conf.editor.visible_fields)
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)

class ToggleEndColumnAction(gaupol.ToggleAction):
    def __new__(cls):
        action = gaupol.ToggleAction.new("toggle-end-column")
        action.__class__ = cls
        return action
    def __init__(self):
        gaupol.ToggleAction.__init__(self, "toggle-end-column")
        self.action_group = "unsafe"
        self.set_state(gaupol.fields.END in gaupol.conf.editor.visible_fields)
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)

class ToggleMainTextColumnAction(gaupol.ToggleAction):
    def __new__(cls):
        action = gaupol.ToggleAction.new("toggle-main-text-column")
        action.__class__ = cls
        return action
    def __init__(self):
        gaupol.ToggleAction.__init__(self, "toggle-main-text-column")
        self.action_group = "unsafe"
        self.set_state(gaupol.fields.MAIN_TEXT in gaupol.conf.editor.visible_fields)
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)

class ToggleMainToolbarAction(gaupol.ToggleAction):
    def __new__(cls):
        action = gaupol.ToggleAction.new("toggle-main-toolbar")
        action.__class__ = cls
        return action
    def __init__(self):
        gaupol.ToggleAction.__init__(self, "toggle-main-toolbar")
        self.action_group = "safe"
        self.set_state(gaupol.conf.application_window.show_main_toolbar)

class ToggleNumberColumnAction(gaupol.ToggleAction):
    def __new__(cls):
        action = gaupol.ToggleAction.new("toggle-number-column")
        action.__class__ = cls
        return action
    def __init__(self):
        gaupol.ToggleAction.__init__(self, "toggle-number-column")
        self.action_group = "unsafe"
        self.set_state(gaupol.fields.NUMBER in gaupol.conf.editor.visible_fields)
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)

class TogglePlayerAction(gaupol.ToggleAction):
    def __new__(cls):
        action = gaupol.ToggleAction.new("toggle-player")
        action.__class__ = cls
        return action
    def __init__(self):
        gaupol.ToggleAction.__init__(self, "toggle-player")
        self.action_group = "safe"
        self.set_state(False)
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(application.player is not None)

class ToggleStartColumnAction(gaupol.ToggleAction):
    def __new__(cls):
        action = gaupol.ToggleAction.new("toggle-start-column")
        action.__class__ = cls
        return action
    def __init__(self):
        gaupol.ToggleAction.__init__(self, "toggle-start-column")
        self.action_group = "unsafe"
        self.set_state(gaupol.fields.START in gaupol.conf.editor.visible_fields)
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)

class ToggleTranslationTextColumnAction(gaupol.ToggleAction):
    def __new__(cls):
        action = gaupol.ToggleAction.new("toggle-translation-text-column")
        action.__class__ = cls
        return action
    def __init__(self):
        gaupol.ToggleAction.__init__(self, "toggle-translation-text-column")
        self.action_group = "unsafe"
        self.set_state(gaupol.fields.TRAN_TEXT in gaupol.conf.editor.visible_fields)
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)

__all__ = tuple(x for x in dir() if x.endswith("Action"))



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Previewing subtitles with a video player."""

import aeidon
import os
import string


class PreviewAgent(aeidon.Delegate):

    """Previewing subtitles with a video player."""

    def __init__(self, master):
        """Initialize a :class:`aeidon.PreviewAgent` instance."""
        aeidon.Delegate.__init__(self, master)
        aeidon.util.connect(self, self, "notify::main_file")

    @aeidon.deco.export
    def find_video(self):
        """
        Find and return the video file path based on main file's path.

        The video file is searched for in the same directory as the subtitle
        file. The subtitle file's filename without extension is assumed to
        start with or match the video file's filename without extension,
        e.g. 'movie.avi' for 'movie.en.srt'.
        """
        if self.main_file is None: return None
        dirname = os.path.dirname(self.main_file.path)
        subname = os.path.basename(self.main_file.path)
        for name in os.listdir(dirname):
            path = os.path.join(dirname, name)
            root = os.path.splitext(name)[0]
            if not subname.startswith(root): continue
            if aeidon.util.is_video_file(path):
                self.video_path = path
                return self.video_path
        return None

    def _get_subtitle_path(self, doc, encoding=None, temp=False):
        """
        Return path to a file to preview, either real or temporary.

        Raise :exc:`IOError` if writing to temporary file fails.
        Raise :exc:`UnicodeError` if encoding temporary file fails.
        """
        file = self.get_file(doc)
        if file is None or encoding != file.encoding:
            return self.new_temp_file(doc)
        if doc == aeidon.documents.MAIN:
            if not self.main_changed and not temp:
                return self.main_file.path
        if doc == aeidon.documents.TRAN:
            if not self.tran_changed and not temp:
                return self.tran_file.path
        return self.new_temp_file(doc)

    def _on_notify_main_file(self, *args):
        """Try to find the video file path if unset."""
        if self.video_path is None:
            self.find_video()

    @aeidon.deco.export
    def preview(self, position, doc, command, offset, encoding=None, temp=False):
        """
        Start video player with `command` from `position`.

        `command` can have variables ``$MILLISECONDS``, ``$SECONDS``,
        ``$SUBFILE`` and ``$VIDEOFILE``. `offset` should be the amount
        of seconds before `position` to start. `encoding` can be specified
        if different from `doc` file encoding. Use ``True`` for `temp` to
        always use a temporary file for preview regardless of whether
        the file is changed or not.

        Return a three tuple of :class:`subprocess.POpen` instance, command
        with variables expanded and a file object to which process standard
        output and standard error are directed.

        Raise :exc:`IOError` if writing to temporary file fails.
        Raise :exc:`UnicodeError` if encoding temporary file fails.
        Raise :exc:`aeidon.ProcessError` if unable to start process.
        """
        sub_path = self._get_subtitle_path(doc, encoding, temp=temp)
        fout = open(aeidon.temp.create(".output"), "w")
        seconds = max(0, self.calc.to_seconds(position) - offset)
        command = string.Template(command).safe_substitute(
            MILLISECONDS=("{:.0f}".format(seconds * 1000)),
            SECONDS=("{:.3f}".format(seconds)),
            SUBFILE=aeidon.util.shell_quote(sub_path),
            VIDEOFILE=aeidon.util.shell_quote(self.video_path))

        process = aeidon.util.start_process(command, stderr=fout, stdout=fout)
        return process, command, fout



# -*- coding: utf-8 -*-

# Copyright (C) 2012 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""GStreamer video player."""

import aeidon
import gaupol
import time

from collections import namedtuple
from aeidon.i18n   import _
from gi.repository import GLib
from gi.repository import Gtk

with aeidon.util.silent(Exception):
    from gi.repository import Gst

__all__ = ("VideoPlayer",)


class VideoPlayer(aeidon.Observable):

    """
    GStreamer video player.

    :ivar audio_track: Current audio track as integer
    :ivar calc: The instance of :class:`aeidon.Calculator` used
    :ivar _in_default_segment: ``True`` if in the default playback segment
    :ivar _info: :class:`Gst.DiscovererInfo` from current URI
    :ivar _playbin: GStreamer "playbin" element
    :ivar _prev_state: Previous state of `_playbin`
    :ivar ready: ``True`` if a file is loaded and ready to play
    :ivar subtitle_text: Current text shown in the subtitle overlay
    :ivar subtitle_text_raw: `subtitle_text` before removal of tags
    :ivar _text_overlay: GStreamer "textoverlay" element
    :ivar _time_overlay: GStreamer "timeoverlay" element
    :ivar volume: Current audio stream volume
    :ivar widget: :class:`Gtk.DrawingArea` used to render video

    Signals and their arguments for callback functions:
     * ``state-changed``: player new state
    """

    signals = ("state-changed",)

    TrackInfo = namedtuple("TrackInfo", ["title", "language_code", "language_name"])

    def __init__(self):
        """Initialize a :class:`VideoPlayer` instance."""
        aeidon.Observable.__init__(self)
        self.calc = aeidon.Calculator()
        self._in_default_segment = True
        self._info = None
        self._playbin = None
        self._prev_state = None
        self.ready = False
        self.subtitle_text_raw = ""
        self._text_overlay = None
        self._time_overlay = None
        self.widget = None
        self._init_widget()
        self._init_text_overlay()
        self._init_time_overlay()
        self._init_pipeline()
        self._init_bus()

    @property
    def audio_track(self):
        """Return number of the current audio track."""
        track = self._playbin.props.current_audio
        # If at the default value, the first track is used,
        # which is (probably?) zero.
        return 0 if track == -1 else track

    @audio_track.setter
    def audio_track(self, track):
        """Set the current audio track."""
        self._playbin.props.current_audio = track

    def _ensure_default_segment(self):
        """
        Reset playback start and stop positions.

        When using `play_segment`, the playback limits are set to those
        (instead of the default 0 and -1). To be able to do useful things,
        we often need to reset those segment limits to their defaults.
        Any methods that need to rely on the whole duration being
        accessible, should as first thing call this method.
        """
        if self._in_default_segment: return
        # XXX: There's got to be a simpler way to do this.
        seek_flags = Gst.SeekFlags.FLUSH | Gst.SeekFlags.ACCURATE
        pos = self.get_position(aeidon.modes.SECONDS) * Gst.SECOND
        end = self.get_duration(aeidon.modes.SECONDS) * Gst.SECOND
        if pos is None: return
        if end is None: return
        self._playbin.seek(rate=1.0,
                           format=Gst.Format.TIME,
                           flags=seek_flags,
                           start_type=Gst.SeekType.SET,
                           start=0,
                           stop_type=Gst.SeekType.SET,
                           stop=end)

        self._playbin.seek_simple(Gst.Format.TIME, seek_flags, pos)
        self._in_default_segment = True

    def get_audio_infos(self):
        """Return a sequence of audio track infos."""
        return tuple(
            self._make_track_infos(self._playbin.emit("get-audio-tags", i))
            for i in range(self._playbin.props.n_audio)
        )

    def _make_track_infos(self, tags):
        return self.TrackInfo(
            tags.get_string("title")[1],
            tags.get_string("language-code")[1],
            tags.get_string("language-name")[1]
        )

    def get_duration(self, mode=None):
        """Return duration of video stream or ``None``."""
        q = self._playbin.query_duration
        for i in range(100):
            # Querying duration sometimes fails very unreproducibly,
            # likely due to a particular pipeline state or state change.
            # Try repeated times, hoping to pass the bad state.
            success, duration = q(Gst.Format.TIME)
            if success: break
            time.sleep(1/100)
        if not success: return None
        if mode is None: return duration
        duration = duration / Gst.SECOND
        if mode == aeidon.modes.SECONDS:
            return duration
        if mode == aeidon.modes.TIME:
            return self.calc.to_time(duration)
        if mode == aeidon.modes.FRAME:
            return self.calc.to_frame(duration)
        raise ValueError("Invalid mode: {!r}".format(mode))

    def get_position(self, mode=None):
        """Return current position in video stream or ``None``."""
        q = self._playbin.query_position
        for i in range(100):
            # Querying position sometimes fails very unreproducibly,
            # likely due to a particular pipeline state or state change.
            # Try repeated times, hoping to pass the bad state.
            success, pos = q(Gst.Format.TIME)
            if success: break
            time.sleep(1/100)
        if not success: return None
        if mode is None: return pos
        pos = pos / Gst.SECOND
        if mode == aeidon.modes.SECONDS:
            return pos
        if mode == aeidon.modes.TIME:
            return self.calc.to_time(pos)
        if mode == aeidon.modes.FRAME:
            return self.calc.to_frame(pos)
        raise ValueError("Invalid mode: {!r}".format(mode))

    def _init_bus(self):
        """Initialize the GStreamer message bus."""
        bus = self._playbin.get_bus()
        bus.add_signal_watch()
        bus.connect("message::eos", self._on_bus_message_eos)
        bus.connect("message::error", self._on_bus_message_error)
        bus.connect("message::state-changed", self._on_bus_message_state_changed)

    def _init_pipeline(self):
        """Initialize the GStreamer playback pipeline."""
        self._playbin = Gst.ElementFactory.make("playbin", None)
        if gaupol.conf.video_player.volume is not None:
            self.volume = gaupol.conf.video_player.volume
        sink = Gst.ElementFactory.make("gtksink", None)
        self.widget.pack_start(sink.props.widget, True, True, 0)
        bin = Gst.Bin()
        bin.add(self._time_overlay)
        bin.add(self._text_overlay)
        pad = self._time_overlay.get_static_pad("video_sink")
        bin.add_pad(Gst.GhostPad.new("sink", pad))
        bin.add(sink)
        self._time_overlay.link(self._text_overlay)
        self._text_overlay.link(sink)
        self._playbin.props.video_sink = bin
        # We need to disable playbin's own subtitle rendering, since we don't
        # want embedded subtitles to be displayed, but rather what we
        # explicitly set to our own overlays. Since Gst.PlayFlags is not
        # available via introspection, we need to use Gst.util_set_object_arg.
        # Playbin's default values can be found via 'gst-inspect-1.0 playbin'.
        Gst.util_set_object_arg(self._playbin,
                                "flags",
                                "+".join(("soft-colorbalance",
                                          "deinterlace",
                                          "soft-volume",
                                          "audio",
                                          "video")))

    def _init_text_overlay(self):
        """Initialize the text overlay element."""
        self._text_overlay = Gst.ElementFactory.make("textoverlay", None)
        conf = gaupol.conf.video_player
        callback = self._on_conf_notify_subtitle_property
        conf.connect("notify::subtitle_font", callback)
        conf.connect("notify::subtitle_color", callback)
        conf.connect("notify::subtitle_alpha", callback)
        conf.connect("notify::subtitle_background", callback)
        self._on_conf_notify_subtitle_property()

    def _init_time_overlay(self):
        """Initialize the time overlay element."""
        self._time_overlay = Gst.ElementFactory.make("timeoverlay", None)
        conf = gaupol.conf.video_player
        callback = self._on_conf_notify_time_property
        conf.connect("notify::time_font", callback)
        conf.connect("notify::time_color", callback)
        conf.connect("notify::time_alpha", callback)
        conf.connect("notify::time_background", callback)
        self._on_conf_notify_time_property()

    def _init_widget(self):
        """Initialize the rendering widget."""
        self.widget = Gtk.Box()
        style = self.widget.get_style_context()
        style.add_class("gaupol-video-background")
        gaupol.style.load_css(self.widget)

    def is_playing(self):
        """Return ``True`` if playing video."""
        # GStreamer's state information is far too detailed for our purposes.
        # Let's consider the state to be playing also if undergoing a state
        # change and/or having a pending playing state.
        success, current, pending = self._playbin.get_state(timeout=1)
        return (success == Gst.StateChangeReturn.ASYNC or
                current == Gst.State.PLAYING or
                pending == Gst.State.PLAYING)

    def _on_bus_message_eos(self, bus, message):
        """Handle EOS message from the bus."""
        self._ensure_default_segment()
        self.pause()

    def _on_bus_message_error(self, bus, message):
        """Handle error message from the bus."""
        title, message = message.parse_error()
        dialog = gaupol.ErrorDialog(None, title, message)
        dialog.add_button(_("_OK"), Gtk.ResponseType.OK)
        dialog.set_default_response(Gtk.ResponseType.OK)
        gaupol.util.flash_dialog(dialog)

    def _on_bus_message_state_changed(self, bus, message):
        """Emit signal if state changed in a relevant manner."""
        old, new, pending = message.parse_state_changed()
        states = (Gst.State.NULL, Gst.State.PAUSED, Gst.State.PLAYING)
        if not new in states: return
        if new == self._prev_state: return
        self.emit("state-changed", new)
        self._prev_state = new

    def _on_conf_notify_subtitle_property(self, *args):
        """Update subtitle text overlay properties."""
        conf = gaupol.conf.video_player
        self._text_overlay.props.font_desc = conf.subtitle_font
        self._text_overlay.props.halignment = conf.subtitle_position_horizontal
        self._text_overlay.props.valignment = conf.subtitle_position_vertical
        self._text_overlay.props.line_alignment = conf.line_alignment
        self._text_overlay.props.shaded_background = conf.subtitle_background
        alpha = "{:02x}".format(int(conf.subtitle_alpha * 255))
        color = conf.subtitle_color.replace("#", "")
        color = int(float.fromhex("".join((alpha, color))))
        self._text_overlay.props.color = color

    def _on_conf_notify_time_property(self, *args):
        """Update time overlay properties."""
        conf = gaupol.conf.video_player
        self._time_overlay.props.font_desc = conf.time_font
        self._time_overlay.props.halignment = conf.time_position_horizontal
        self._time_overlay.props.valignment = conf.time_position_vertical
        self._time_overlay.props.shaded_background = conf.time_background
        alpha = "{:02x}".format(int(conf.time_alpha * 255))
        color = conf.time_color.replace("#", "")
        color = int(float.fromhex("".join((alpha, color))))
        self._time_overlay.props.color = color

    def pause(self):
        """Pause."""
        self._playbin.set_state(Gst.State.PAUSED)

    def play(self):
        """Play."""
        self._playbin.set_state(Gst.State.PLAYING)

    def play_segment(self, start, end):
        """Play from `start` to `end`."""
        self._in_default_segment = False
        seek_flags = Gst.SeekFlags.FLUSH | Gst.SeekFlags.ACCURATE
        start = max(0, self.calc.to_seconds(start)) * Gst.SECOND
        duration = self.get_duration(aeidon.modes.SECONDS)
        if duration is None: return
        end = min(duration, self.calc.to_seconds(end)) * Gst.SECOND
        self._playbin.seek(rate=1.0,
                           format=Gst.Format.TIME,
                           flags=seek_flags,
                           start_type=Gst.SeekType.SET,
                           start=start,
                           stop_type=Gst.SeekType.SET,
                           stop=end)

        self.play()

    def seek(self, pos):
        """Seek to `pos`."""
        self._ensure_default_segment()
        seek_flags = Gst.SeekFlags.FLUSH | Gst.SeekFlags.ACCURATE
        pos = self.calc.to_seconds(pos) * Gst.SECOND
        self._playbin.seek_simple(Gst.Format.TIME, seek_flags, pos)

    def _seek_null(self):
        """Seek to the current position."""
        self._ensure_default_segment()
        pos = self.get_position(mode=None)
        if pos is None: return
        seek_flags = Gst.SeekFlags.FLUSH | Gst.SeekFlags.ACCURATE
        self._playbin.seek_simple(Gst.Format.TIME, seek_flags, pos)

    def seek_relative(self, offset):
        """Seek to `offset` relative to current position."""
        self._ensure_default_segment()
        pos = self.get_position(aeidon.modes.SECONDS)
        duration = self.get_duration(aeidon.modes.SECONDS)
        if pos is None or duration is None: return
        pos = pos + self.calc.to_seconds(offset)
        pos = max(0, min(pos, duration))
        pos = pos * Gst.SECOND
        seek_flags = Gst.SeekFlags.FLUSH | Gst.SeekFlags.ACCURATE
        self._playbin.seek_simple(Gst.Format.TIME, seek_flags, pos)

    def set_path(self, path):
        """Set the path of the file to play."""
        self.set_uri(aeidon.util.path_to_uri(path))

    def set_uri(self, uri):
        """Set the URI of the file to play."""
        self.ready = False
        self._playbin.props.uri = uri
        self.subtitle_text = ""
        try:
            # Find out the exact framerate to be able
            # to convert between position types.
            from gi.repository import GstPbutils
            discoverer = GstPbutils.Discoverer()
            self._info = discoverer.discover_uri(uri)
            streams = self._info.get_stream_list()
            stream_types = [x.get_stream_type_nick() for x in streams]
            if "video" not in stream_types:
                raise Exception(_("No video streams found"))
            stream = self._info.get_video_streams()[0]
            num = float(stream.get_framerate_num())
            denom = float(stream.get_framerate_denom())
            self.calc = aeidon.Calculator(num/denom)
            self.ready = True
        except Exception as error:
            title = _("Failed to initialize playback")
            dialog = gaupol.ErrorDialog(None, title, str(error))
            dialog.add_button(_("_OK"), Gtk.ResponseType.OK)
            dialog.set_default_response(Gtk.ResponseType.OK)
            gaupol.util.flash_dialog(dialog)
        else:
            # Make stream tags available from _playbin
            self._playbin.set_state(Gst.State.PAUSED)
            self._playbin.get_state(Gst.CLOCK_TIME_NONE)

    def stop(self):
        """Stop."""
        self._playbin.set_state(Gst.State.NULL)
        self.subtitle_text = ""

    @property
    def subtitle_text(self):
        """Return text shown in the subtitle overlay."""
        return self._text_overlay.props.text

    @subtitle_text.setter
    def subtitle_text(self, text):
        """Set `text` to the subtitle overlay."""
        self.subtitle_text_raw = text
        text = aeidon.RE_ANY_TAG.sub("", text)
        text = GLib.markup_escape_text(text)
        self._text_overlay.props.text = text
        # Do a zero seek to force an update of the overlay.
        # https://github.com/otsaloma/gaupol/issues/181
        if self._in_default_segment:
            with aeidon.util.silent(Exception, tb=True):
                self._seek_null()

    @property
    def volume(self):
        """Return the current volume, in range [0,1]."""
        return self._playbin.props.volume

    @volume.setter
    def volume(self, volume):
        """Set the volume to use, in range [0,1]."""
        volume = round(max(0, min(1, volume)), 2)
        if abs(volume - self.volume) < 0.001: return
        self._playbin.props.volume = volume
        gaupol.conf.video_player.volume = volume



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""GTK user interface for the Gaupol subtitle editor."""

__version__ = "1.10.1"
COMBO_SEPARATOR = "<separator/>"

import sys
import warnings

if hasattr(sys, "frozen"):
    # Avoid error trying to write to non-existent stderr.
    # https://stackoverflow.com/a/35773092
    warnings.simplefilter("ignore")

# Needed for the spell-check dialog, where the '_proceed' function call is
# recursive. The recursion limit effectively defines the maximum allowed
# amount of consecutive subtitles without spelling errors, which if exceeded
# triggers a RecursionError.
sys.setrecursionlimit(10000)

import aeidon
import gi
import os

# Disable gst-vaapi as it doesn't seem to work with gtksink.
# https://github.com/otsaloma/gaupol/issues/79
os.environ["LIBVA_DRIVER_NAME"] = "null"
os.environ["LIBVA_DRIVERS_PATH"] = "/dev/null"

gi.require_version("Gdk", "3.0")
gi.require_version("Gtk", "3.0")

for module, version in {
    "Gst": "1.0",
    "GstPbutils": "1.0",
    "GstVideo": "1.0",
    "GstTag": "1.0",
}.items():
    with aeidon.util.silent(Exception):
        gi.require_version(module, version)

with aeidon.util.silent(Exception):
    from gi.repository import Gst
    Gst.init(None)

from gaupol.urls import * # noqa
from gaupol import util # noqa
from gaupol.enums import * # noqa
from gaupol.errors import * # noqa
from gaupol.attrdict import * # noqa
from gaupol.config import ConfigurationStore # noqa
conf = ConfigurationStore()
from gaupol import style # noqa
from gaupol import ruler # noqa
from gaupol.entries import * # noqa
from gaupol.renderers import * # noqa
from gaupol.floatlabel import * # noqa
from gaupol.spell import * # noqa
from gaupol.view import * # noqa
from gaupol.page import * # noqa
from gaupol.player import * # noqa
from gaupol.dialogs.builder import * # noqa
from gaupol.dialogs.file import * # noqa
from gaupol.dialogs.open import * # noqa
from gaupol.dialogs.save import * # noqa
from gaupol.dialogs.video import * # noqa
from gaupol.dialogs.append import * # noqa
from gaupol.dialogs.about import * # noqa
from gaupol.dialogs.debug import * # noqa
from gaupol.dialogs.duration_adjust import * # noqa
from gaupol.dialogs.encoding import * # noqa
from gaupol.dialogs.framerate_convert import * # noqa
from gaupol.dialogs.insert import * # noqa
from gaupol.dialogs.language import * # noqa
from gaupol.dialogs.message import * # noqa
from gaupol.dialogs.multi_close import * # noqa
from gaupol.dialogs.multi_save import * # noqa
from gaupol.dialogs.preferences import * # noqa
from gaupol.dialogs.preview_error import * # noqa
from gaupol.dialogs.search import * # noqa
from gaupol.dialogs.position_shift import * # noqa
from gaupol.dialogs.spell_check import * # noqa
from gaupol.dialogs.split import * # noqa
from gaupol.dialogs.text_edit import * # noqa
from gaupol.dialogs.position_transform import * # noqa
from gaupol.assistants import * # noqa
from gaupol.action import * # noqa
from gaupol import actions # noqa
from gaupol.extension import * # noqa
from gaupol.extensionman import * # noqa
from gaupol import agents # noqa
from gaupol.application import * # noqa
from gaupol.applicationman import ApplicationManager # noqa
from gaupol.unittest import * # noqa

def main(args):
    """Initialize application."""
    global appman
    aeidon.i18n.bind()
    appman = ApplicationManager(args)
    raise SystemExit(appman.run())



# -*- coding: utf-8 -*-

# Copyright (C) 2012 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Video actions for :class:`gaupol.Application`."""

import aeidon
import gaupol

class LoadVideoAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "load-video")
        self.accelerators = ["<Control>L"]
        self.action_group = "safe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(gaupol.util.gst_available())
        aeidon.util.affirm(page is not None)

class PlayPauseAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "play-pause")
        self.accelerators = ["P"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(application.player is not None)
        aeidon.util.affirm(application.player.ready)

class PlaySelectionAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "play-selection")
        self.accelerators = ["O"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(application.player is not None)
        aeidon.util.affirm(application.player.ready)
        aeidon.util.affirm(selected_rows)

class SeekBackwardAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "seek-backward")
        self.accelerators = ["<Shift><Control>Left"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(application.player is not None)
        aeidon.util.affirm(application.player.ready)

class SeekForwardAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "seek-forward")
        self.accelerators = ["<Shift><Control>Right"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(application.player is not None)
        aeidon.util.affirm(application.player.ready)

class SeekNextAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "seek-next")
        self.accelerators = ["<Control>Right"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(application.player is not None)
        aeidon.util.affirm(application.player.ready)

class SeekPreviousAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "seek-previous")
        self.accelerators = ["<Control>Left"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(application.player is not None)
        aeidon.util.affirm(application.player.ready)

class SeekSelectionEndAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "seek-selection-end")
        self.accelerators = ["<Control>Down"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(application.player is not None)
        aeidon.util.affirm(application.player.ready)
        aeidon.util.affirm(selected_rows)

class SeekSelectionStartAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "seek-selection-start")
        self.accelerators = ["<Control>Up"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(application.player is not None)
        aeidon.util.affirm(application.player.ready)
        aeidon.util.affirm(selected_rows)

__all__ = tuple(x for x in dir() if x.endswith("Action"))
